# editable.rb
#
# Contributors:: Oliver Smith


module Mondrian

    # A mixin which implements the clean/dirty pattern.
    #
    # Adds two events, +clean+ and +dirty+, to the receiver.  These events are
    # triggered when the receiver is modified and calls the +dirty+ or +clean+ methods.
    # Also keeps track of the current dirty/clean state by methods +is_dirty?+ and
    # +is_clean?+.
    
    module Editable
        
        def self.append_features(base)
            super
            
            base.class_eval <<-fin
                require 'framework/mixins/notifier'
            
                include Notifier
                
                event :dirty, :clean
            fin
        end
    
    
        def is_dirty?
            (defined? @editable_is_dirty) ? @editable_is_dirty : false
        end
        
        
        def is_clean?
            (defined? @editable_is_dirty) ? (not @editable_is_dirty) : true
        end
        

        # Called by the receiver when it has been modified.
    
        protected
        def dirty
            return false if (defined? @editable_is_dirty) and @editable_is_dirty
        
            @editable_is_dirty = true
        
            notify_dirty
            
            true
        end
        
        
        # Called by the receiver when it has been saved or otherwise cleaned.
        
        protected
        def clean
            return false if (not defined? @editable_is_dirty) or (not @editable_is_dirty)
            
            @editable_is_dirty = false
            
            notify_clean
            
            true
        end
        
    end
    
end