# notifier.rb
#
# Contributors:: Oliver Smith


module Mondrian

    # A mixin which implements the notify/subscribe pattern.
    #
    # Enables objects to receive notification from other objects by
    # subscribing to specific events implemented by the Notifier object.
    #
    # Notes:: The downside to using symbols to refer to events is that they
    #         are not checked by the syntax checker.  However, if a subscriber
    #         uses separate methods or blocks for each event, it should never
    #         be necessary to refer to an event by its symbol.
    
    module Notifier
    
        module ClassMethods
        
            # Adds a list of events to this object.
            #
            # For each symbol, adds the methods +subscribe_symbol+, 
            # +unsubscribe_symbol+ and +notify_symbol+ to this object.
            
            def event(*symbols)
                symbols.each { |symbol| 
                    class_eval <<-fin
                        def subscribe_#{symbol}(subscriber, callback = nil, &block)
                            subscribe(subscriber, :#{symbol}, callback || block)
                        end
                        
                        
                        def unsubscribe_#{symbol}(subscriber)
                            unsubscribe(subscriber, :#{symbol})
                        end


                        protected
                        def notify_#{symbol}(*args)
                            notify(:#{symbol}, *args)
                        end
                    fin
                }
            end
        end


        def self.append_features(base)
            super
            
            base.extend ClassMethods
        end


        # Allows +subscriber+ to receive notifications when +event+ occurs.
        #
        # The +callback+ method or +block+ will be called with three arguments:
        # this object, the event triggered and arbitrary arguments provided
        # by the notifier object.
        
        def subscribe(subscriber, event, callback = nil, &block)
            raise "No callback or block provided!" unless callback or block
        
            subscribers[subscriber] ||= {}

            subscribers[subscriber][event] = callback or block
            
            true
        end
        
        
        # Allows +subscriber+ to receive notifications of all events from this object.
        #
        # Uses a special :all symbol to denote that all events are handled.  Additional
        # handlers may be added after this statement and will coexist.
        
        def subscribe_all(subscriber, callback = nil, &block)
            subscribers[subscriber] = { :all => callback || block }
        end
        
        
        # Unsubscribes +subscriber+ from receiving notifications of +event+.
        
        def unsubscribe(subscriber, event)
            subscribers[subscriber].delete event
            
            true
        end
        
        
        # Unsubscribes +subscriber+ from all events triggered by this object.
        
        def unsubscribe_all(subscriber)
            subscribers.delete(subscriber)
            
            true
        end
    
    
        protected
        def subscribers
            @subscribers = {} unless defined? @subscribers
            
            @subscribers
        end
        

        # Notifies any subscribers that +event+ has occurred.
        #
        # Passes +args+ along to the subscriber.

        protected
        def notify(event, *args)
            args = nil if args.size == 0
        
            subscribers.each { |subscriber, events|
                if events.include? event
                    events[event].call(self, event, *args) 
                    
                elsif events.include? :all
                    events[:all].call(self, event, *args) 
                end
            }
        end
    end
end



