# project.rb
#
# Contributors:: Oliver Smith


require 'yaml'

require 'framework/common/utilities'

include Mondrian


module Mondrian

    # A class for managing a project directory.

    class Project
        attr_reader :name, :path
        
        ProjectFilename = 'project.yml'
        
        # Creates a new project at +path+, creating the folder if necessary.
        
        def self.create(path, name)
            Dir.mkdir(path) unless File.exists?(path)
            
            raise Errno::EEXIST if File.exists?(File.join(path, ProjectFilename))
            
            File.open(File.join(path, ProjectFilename), 'w') { |file|
                file.write({ 'Project' => name }.to_yaml)
            }
            
            self.new(path)
        end
        
        
        # Returns whether +abs_path+ is a project folder.

        def self.project_folder?(abs_path)
            raise ArgumentError, "Path is not absolute!" unless is_absolute? abs_path

            return is_real_directory?(abs_path) && File.exists?(File.join(abs_path, ProjectFilename))
        end


        # Where +path+ is a directory which contains a +project.yml+ file.
        
        def initialize(path)
            raise Errno::ENOENT unless File.directory?(path)
            raise Errno::ENOENT unless File.exists?(File.join(path, ProjectFilename))
        
            @path = File.expand_path(path)
            
            configure
        end
        
        
        # Returns whether +abs_path+ is a non-linked folder in this project.

        def project_subfolder?(abs_path)
            raise ArgumentError, "Path (#{abs_path}) is not absolute!" unless is_absolute? abs_path

            return is_real_directory?(abs_path) && is_subpath?(abs_path, @path)
        end
        
        
        # Returns whether +abs_path+ is a file contained in this project.

        def project_file?(abs_path)
            raise ArgumentError, "Path (#{abs_path}) is not absolute!" unless is_absolute? abs_path
        
            return File.file?(abs_path) && is_subpath?(abs_path, @path) && File.basename(abs_path) != ProjectFilename
        end

        
        # Returns the absolute names of project subfolders non-recursively.
        
        def folders(abs_path = @path)
            raise ArgumentError, "Path (#{abs_path}) does not refer to a folder in this project!" unless project_subfolder?(abs_path)

            folders = Dir.new(abs_path).entries.collect { |entry| File.join(abs_path, entry) }.select { |entry| project_subfolder? entry }
            
            folders.sort { |a, b| File.basename(a).downcase <=> File.basename(b).downcase }
        end
        

        # Returns the absolute names of project files non-recursively.
        
        def files(abs_path = @path)
            raise ArgumentError, "Path (#{abs_path}) does not refer to a file in this project!" unless project_subfolder?(abs_path)
        
            files = Dir.new(abs_path).entries.collect { |entry| File.join(abs_path, entry) }.select { |entry| project_file? entry }
            
            files.sort { |a, b| File.basename(a).downcase <=> File.basename(b).downcase }
        end


        private

        # Configures this project from the yaml configuration file.
        
        def configure
            config = YAML::load(File.new(File.join(@path, ProjectFilename), 'r'))
            
            @name = config['Project']
        end
        
    end
    
end