# resource.rb
#
# Contributors:: Oliver Smith


module Mondrian

    # A class for representing a resource, such as an icon, which may be
    # associated with a particular instance, such as the specific icon used
    # to represent files of a certain type.  Subclass this and give it whatever 
    # attributes it needs to represent your resource.  Simply instantiating the 
    # subclass will register the instance under the appropriate type.  The base 
    # class holds a list of all registered +Resource+ types and instances in 
    # +@resources+.
    
    class Resource
        def Resource.resources
            @@resources = {} unless defined? @@resources
            
            @@resources
        end
    
    
        def Resource.inherited(klass)
            resources[klass] = []
        end
        
        
        # Returns the appropriate +Resource+ for +object+.  The appropriate
        # resource is the last one to have registered which signals capability
        # of handling +object+.
        
        def Resource.get_resource_for(object)
            resource = nil
        
            if resources.include? self
                resources[self].each { |r|
                    resource = r if r.handles? object
                }
            end
            
            resource
        end
        
        
        # +klasses+ is the list of classes this +Resource+ applies to.  
        # +klasses+ may be omitted if +handles_proc+ is provided to evaluate
        # whether an object is handled by this instance or not, or both may
        # be used to narrow down the class and fine detail.
        
        def initialize(klasses = [], &handles_proc)
            @klasses = klasses
            @handles_proc = handles_proc
        
            Resource.resources[self.class] << self
        end
        
        
        # Returns +true+ if this Resource instance is associated with +object+.
        # Children can subclass for more fine-tuned behavior, but the default
        # can check whether +object+'s class is found in +@klasses+, and if 
        # the +@handles_proc+ passed in to +initialize+ evaluates to +true+.
        
        def handles?(object)
            if @handles_proc
                return (@klasses.empty? || @klasses.include?(object.class)) && @handles_proc.call(object)
            else
                return @klasses.include?(object.class)
            end
        end
    end
    
end