# tokenstring.rb
#
# Contributors:: Oliver Smith


require 'framework/common/tokenencoder'
require 'framework/common/utilities'
require 'framework/mixins/notifier'
require 'framework/lexer'

require 'strscan'


module Mondrian

    # A mixin which provides +TokenString+ with various features.

    module TokenStringInternals

        def self.append_features(receiver)
            super
            
            receiver.extend(ClassMethods)
        end
        
        
        def lexer
            self.class.lexer
        end
    
        
        module ClassMethods
            attr_reader :lexer
        
            # Adds methods pertaining to a +Lexer+ to this +TokenString+.
            #
            # For each of the +Lexer+'s tokens, adds the following methods:
            #
            # * +each_token+
            # * +next_token+
            # * +previous_token+
            #
            # For example, if a +Lexer+ had an :operator token, +next_operator+
            # would be added to the +TokenString+.
        
            private
            def use_lexer(lexer_class)
                @lexer = lexer_class.new
        
                (@lexer.tokens + @lexer.groups.keys).each { |symbol|
                    module_eval <<-fin
                        def each_#{symbol}(text_match = nil, &block)
                            each(:#{symbol}, text_match, &block)
                        end

            
                        def next_#{symbol}(text_match = nil, start_position = 0)
                            self.next(:#{symbol}, text_match, start_position)
                        end
            
            
                        def previous_#{symbol}(text_match = nil, start_position = nil)
                            previous(:#{symbol}, text_match, start_position)
                        end
                    fin
                }
            end

        end

    end
        


    # A class for working with a specific token in a +TokenString+.

    class TokenPointer
        attr_reader :position, :text, :token_type, :parent
        
        include Comparable
        
        def initialize(parent, position, length)
            @parent     = parent
            @position   = position
            
            @text       = parent.text[position...position + length]
            @token_type = parent.encoder.token_symbol(parent.tokens[position...position + length])
            
            @change = parent.changelog.length - 1
        end
        
        
        # Compares +object+ to this +TokenPointer+'s position.
        #
        # +object+ is a +TokenPointer+ or a +Fixnum+ representing a position
        # in the parent +TokenString+.
        
        def <=>(object)
            if object.kind_of? self.class
                return position.send('<=>', object.position)
            elsif object.kind_of? Fixnum
                return position.send('<=>', object)
            end
            
            raise TypeError
        end
        
        
        # Compares equality to a variety of objects.
        #
        # If +object+ is a +Fixnum+, compares to this +TokenPointer+'s position.
        # If +object+ is a +TokenPointer+, compares whether it points to the same token.
        # If +object+ is a token symbol, compares whether +token_type+ is the same.
        # If +object+ is a +String+, compares whether the token text is the same.
        
        def ==(object)
            if object.kind_of? String
                return text == object
            elsif object.kind_of? self.class
                return (object.text == text and object.token_type == token_type and object.position == position)
            elsif object.kind_of? Symbol
                return parent.lexer.in_group?(token_type, object) if parent.lexer.is_group?(object)
                return token_type == object
            elsif object.kind_of? Fixnum
                return position == object
            end
            
            super
        end
        
        
        # A synonym for +text+.
        
        def to_s
            text
        end
        
        
        # Returns the length of the token pointed to.
        
        def length
            text.length
        end
        
        
        # Returns this token's starting index in the parent +TokenString+.
        
        def position
            update_position
            
            @position
        end
        
        
        # Corrects +position+ by looking at changes made to +parent+.
        
        def update_position
            return if @change == @parent.changelog.length - 1
        
            @parent.changelog[@change + 1...@parent.changelog.length].each { |change|
                @position += change[1] if @position >= change[0]
            }
            
            @change = @parent.changelog.length - 1
        end
        
        
        # Returns whether or not the token pointed to is still intact.
        
        def deceased?
            !(self == @parent.token_at(position))
        end
        

        # Asserts that the token pointed to must still be intact.
        
        def assert_untainted
            raise "Object has expired!" if deceased?
        end


        # Allows the caller to perform token methods as with a +TokenString+.
        #
        # The suite of token methods such as +next_token+ and +previous_token+
        # work on a +TokenPointer+ as well as a +TokenString+.

        def method_missing(method_symbol, *args)
            if method_symbol.to_s =~ /\A(next|previous)_?/
                case $&
                    when 'next'
                        return @parent.send(method_symbol, args[0], args[1], position + length) 
                    when 'previous'
                        return @parent.send(method_symbol, args[0], args[1], position - 1) 
                    when 'next_'
                        return @parent.send(method_symbol, args[0], position + length) 
                    when 'previous_'
                        return @parent.send(method_symbol, args[0], position - 1) 
                end
            end
            
            raise NoMethodError, method_symbol.to_s
        end
    end



    # Contains a string of code operated on by a +Lexer+.
    # 
    # Aims to provide a simple interface for manipulating a code string
    # which can be lexed by a +Lexer+.  Children subclass to specify
    # which +Lexer+ to use.
    #
    # The code is represented by two strings of equal length.  The
    # first string is simply the original text.  The second string is
    # a layer of characters which identify where tokens begin and end and
    # their types.  For example:
    #
    #  my_var = 5
    #  1~~~~~0203
    #
    # The encoded token string contains 4 tokens.  +my_var+ is represented
    # by the text '1~~~~~', the space is represented by the text '0', the
    # operator token '=' is '2', another space is another '0' and the final number
    # token is '3'.  If a token spans more than one character, as in the
    # +my_var+ case, a pad character '~' is used.  This is an easy way of
    # storing tokens which is simple to manually inspect and allows all kinds
    # of regexp fun, which makes it fast to work with.  It's also trivial to find
    # the token at a certain index.
    #
    # A +TokenEncoder+ is used to take care of the encoding details.
    #
    # Notes:: +TokenString+ keeps a log of text inserted or deleted in +changelog+.
    #         This is simply a list of +Range+s.  If 5 characters are inserted at
    #         index 7, +changelog+ will contain (7, 5).  If 5 characters are deleted,
    #         +changelog+ will contain (7, -5).  This allows a +TokenPointer+ to keep
    #         track of a chunk of text even while it is being edited.

    class TokenString
        include Enumerable
        include Notifier
        include TokenStringInternals
        
        attr_reader :text, :tokens, :encoder, :changelog
        
        
        def initialize(text = '')
            @encoder = TokenEncoder.new(lexer.tokens, lexer.groups)
            
            @text        = ""
            @tokens      = ""
            
            @changelog = []
           
            insert_text(0, text)
        end
        
        
        # Implements the equality operator for +String+s and +TokenString+s.
        
        def ==(object)
            if object.kind_of? String
                return @text == object
            elsif object.kind_of? self.class
                return object_id == object.object_id
            end
            
            false
        end
        
        
        # Replaces the current string with new code.
        
        def text=(text)
            delete_text 0, @text.length
            insert_text 0, text
        end
        
        
        # Returns the length of the code string.
        
        def length
            @text.length
        end
        
        
        # Inserts +new_text+ at +index+ and re-lexes the changed text.
        #
        # Stores change markers in +changelog+.  Returns a +Range+ of 
        # the text that was re-lexed (which could be more than just the
        # directly changed text).
        
        def insert_text(index, new_text)
            @text[index, 0]   = new_text
            @tokens[index, 0] = @encoder.null_pad new_text.length
            
            @changelog << [ index, new_text.length ]

            lex_text(index, new_text.length)
        end
        
        
        # Deletes +length+ characters at +index+ and re-lexes the changed text.
        #
        # Stores change markers in +changelog+.  Returns a +Range+ of 
        # the text that was re-lexed (which could be more than just the
        # directly changed text).

        def delete_text(index, length)
            @text[index, length]   = ''
            @tokens[index, length] = ''
        
            @changelog << [ index, -length ]

            lex_text(index)
        end


        def lex_text(index, length = 1)
            lex_start         = get_lex_start(index)
            context           = get_context(lex_start)
            last_change       = index + length - 1
            lex_end           = lex_start
            matched_pos       = 0
            encoded_token     = nil
            last_was_changed  = false
            force_continue    = true
            
            scanner     = StringScanner.new(@text)
            scanner.pos = lex_start

            while !scanner.eos? and (last_was_changed or force_continue)
                matched_pos = scanner.pos
            
                token_type, token_text = lexer.next_token(scanner, context)
                
                encoded_token = @encoder.encode_token(token_type, token_text.length)
                
                last_was_changed = true
                
                if scanner.pos > last_change and @tokens.index(encoder.match_any, matched_pos) == matched_pos
                    if $& == encoded_token and @text.index(token_text, matched_pos) == matched_pos
                        last_was_changed = false
                    end
                end

                if last_was_changed
                    @tokens[matched_pos..scanner.pos - 1] = encoded_token
                    
                    lex_end = scanner.pos - 1
                end
                
                force_continue = context.update!(token_text, token_type)
            end
            
            (lex_start..lex_end)
        end
        
        
        # A synonym for +text+.
        
        def to_s
            text
        end
        
        
        # Returns a +TokenPointer+ for the token at +index+.
        
        def token_at(index)
            return TokenPointer.new(self, $~.begin(0), $&.length) if @tokens.rindex(@encoder.match_any, index)

            nil                
        end
        
        
        # Iterates through each token, returning +TokenPointer+s.
        #
        # +token_type+ can specify to return only tokens of a certain type,
        # and/or +text_match+ is a string or regexp to match within a token's text,
        # e.g. each(:keyword, 'if')
        
        def each(token_type = nil, text_match = nil)
            text_match  = regexp_from_string(text_match) if text_match and text_match.kind_of?(String)
            token_match = token_type ? @encoder.match(token_type) : @encoder.match_any
            
            scanner = StringScanner.new(@tokens)
            
            while scanner.skip_until(token_match)
                if not text_match or @text[scanner.pos - scanner.matched_size...scanner.pos] =~ text_match
                    yield TokenPointer.new(self, scanner.pos - scanner.matched_size, scanner.matched_size)
                end
            end
        end
        
        
        # Returns the token (a +TokenPointer+) meeting the given criteria (see +each+).
        
        def next(token_type = nil, text_match = nil, start_position = 0)
            return nil if start_position < 0 or start_position >= @tokens.length
        
            text_match  = regexp_from_string(text_match) if text_match and text_match.kind_of?(String)
            token_match = token_type ? @encoder.match(token_type) : @encoder.match_any
            
            scanner = StringScanner.new(@tokens)
            scanner.pos = start_position
            
            while scanner.skip_until(token_match)
                if not text_match or @text[scanner.pos - scanner.matched_size...scanner.pos] =~ text_match
                    return TokenPointer.new(self, scanner.pos - scanner.matched_size, scanner.matched_size)
                end
            end
            
            nil
        end
        
        
        # Returns the token (a +TokenPointer+) meeting the given criteria (see +each+).
        
        def previous(token_type = nil, text_match = nil, start_position = nil) # FIXME: needs optimization
            start_position = @tokens.length unless start_position
        
            return nil if start_position <= 0 or start_position > @tokens.length
            
            match = regexp_from_string(match) if text_match and text_match.kind_of?(String)

            token_index = start_position

            while token_index and token_index >= 0
                token_index = @tokens.rindex(@encoder.match(token_type), token_index)
                
                if token_index
                    encoded_token = $&

                    if not text_match or @text[token_index...token_index + encoded_token.length] =~ text_match
                        return TokenPointer.new(self, token_index, encoded_token.length)
                    end
                    
                    token_index -= 1
                end
            end
            
            return nil
        end
        
        
        # Specifies where lexing should begin if a change happens at +index+.
        #
        # If a piece of code is changed, +get_lex_start+ is called with the 
        # starting index of the change.  If a certain language has tokens which
        # are affected by the following token, it should return the index of
        # the previous token to ensure that all tokens that might change are
        # re-lexed.  Children should override this.
        
        protected
        def get_lex_start(change_index) # FIXME: default implementation should move to the start of the current token
            change_index
        end
        
        
        # Gives children a chance to discover the initial context when lexing.
        #
        # Returns a +LexerContext+ appropriate for the given index.  Language
        # implementations should provide a +LexerContext+ that contains all
        # information needed to lex code at a certain index.
        
        protected
        def get_context(index)
            i = @tokens.rindex(@encoder.match_any, index - 1)
            
            last = i ? @text[i, $&.length] : nil
            last_token = last ? @encoder.token_symbol($&) : nil

            LexerContext.new(last, last_token)
        end

    end

end